@@ -409,7 +409,7 @@ void rtdm_task_join_nrt(rtdm_task_t *task, unsigned int poll_delay)
 	int t;
 
 
-	trace_mark(xn_rtdm_task_joinnrt, "thread %p poll_delay %u",
+	trace_mark(xn_rtdm, task_joinnrt, "thread %p poll_delay %u",
 		   task, poll_delay);
 
 	for (t = 0; task->magic && t < JOIN_TIMEOUT; t += poll_delay) {
@@ -425,8 +425,6 @@ void rtdm_task_join_nrt(rtdm_task_t *task, unsigned int poll_delay)
 
 
 
-
-
 EXPORT_SYMBOL(rtdm_task_join_nrt);
 
 /**
@@ -549,11 +547,11 @@ int rtdm_timer_start(rtdm_timer_t *timer, nanosecs_abs_t expiry,
 	int err;
 
 
-
 	err = xntimer_start(timer, xntbase_ns2ticks(rtdm_tbase, expiry),
 			    xntbase_ns2ticks(rtdm_tbase, interval),
 			    (xntmode_t)mode);
 
+
 	return err;
 }
 
@@ -786,7 +784,8 @@ void rtdm_event_init(rtdm_event_t *event, unsigned long pending)
 {
 	spl_t s;
 
-	trace_mark(xn_rtdm_event_init, "event %p pending %lu", event, pending);
+	trace_mark(xn_rtdm, event_init,
+		   "event %p pending %lu", event, pending);
 
 
 
@@ -867,7 +866,7 @@ void rtdm_event_signal(rtdm_event_t *event)
 	unsigned long flags;
 
 
-	trace_mark(xn_rtdm_event_signal, "event %p", event);
+	trace_mark(xn_rtdm, event_signal, "event %p", event);
 
 	flags = rt_global_save_flags_and_cli();
 	__set_bit(0, &event->pending);
@@ -957,6 +956,9 @@ EXPORT_SYMBOL(rtdm_event_wait);
  * - -EPERM @e may be returned if an illegal invocation environment is
  * detected.
  *
+ * - -EWOULDBLOCK is returned if a negative @a timeout (i.e., non-blocking
+ * operation) has been specified.
+ *
  * Environments:
  *
  * This service can be called from:
@@ -975,7 +977,7 @@ int rtdm_event_timedwait(rtdm_event_t *event, nanosecs_rel_t timeout,
 
 
 
-	trace_mark(xn_rtdm_event_timedwait,
+	trace_mark(xn_rtdm, event_timedwait,
 		   "event %p timeout %Lu timeout_seq %p timeout_seq_value %Lu",
 		   event, (long long)timeout, timeout_seq, (long long)(timeout_seq ? *timeout_seq : 0));
 
@@ -1024,6 +1026,7 @@ int rtdm_event_timedwait(rtdm_event_t *event, nanosecs_rel_t timeout,
 
 
 
+
 
 
 EXPORT_SYMBOL(rtdm_event_timedwait);
@@ -1047,7 +1050,8 @@ EXPORT_SYMBOL(rtdm_event_timedwait);
 void rtdm_event_clear(rtdm_event_t *event)
 {
 
-        trace_mark(xn_rtdm_event_clear, "event %p", event);
+
+	trace_mark(xn_rtdm, event_clear, "event %p", event);
 
 
 	event->pending = 0;
@@ -1057,7 +1061,6 @@ void rtdm_event_clear(rtdm_event_t *event)
 
 
 
-
 EXPORT_SYMBOL(rtdm_event_clear);
 
 #ifdef CONFIG_RTAI_RTDM_SELECT
@@ -1076,8 +1079,6 @@ EXPORT_SYMBOL(rtdm_event_clear);
  *
  * @return 0 on success, otherwise:
  *
- * - -EIDRM is returned if @a event has been destroyed.
- *
  * - -ENOMEM is returned if there is insufficient memory to establish the
  * dynamic binding.
  *
@@ -1105,15 +1106,16 @@ int rtdm_event_select_bind(rtdm_event_t *event, rtdm_selector_t *selector,
 		return -ENOMEM;
 
 	xnlock_get_irqsave(&nklock, s);
-	if (unlikely(event->synch_base.magic != RT_SEM_MAGIC))
-		err = -EIDRM;
-	else
-		err = xnselect_bind(&event->select_block,
-				    binding, selector, type, fd_index,
-				    event->pending);
+	err = xnselect_bind(&event->select_block,
+			    binding, selector, type, fd_index,
+			    event->pending || 
+			    event->synch_base.magic != RT_SEM_MAGIC);
 
 	xnlock_put_irqrestore(&nklock, s);
 
+	if (err)
+		xnfree(binding);
+
 	return err;
 }
 EXPORT_SYMBOL(rtdm_event_select_bind);
@@ -1145,7 +1147,7 @@ void rtdm_sem_init(rtdm_sem_t *sem, unsigned long value)
 {
 	spl_t s;
 
-	trace_mark(xn_rtdm_sem_init, "sem %p value %lu", sem, value);
+	trace_mark(xn_rtdm, sem_init, "sem %p value %lu", sem, value);
 
 
 
@@ -1258,7 +1260,8 @@ int rtdm_sem_timeddown(rtdm_sem_t *sem, nanosecs_rel_t timeout,
 
 
 
-	trace_mark(xn_rtdm_sem_timedwait,
+
+	trace_mark(xn_rtdm, sem_timedwait,
 		   "sem %p timeout %Lu timeout_seq %p timeout_seq_value %Lu",
 		   sem, (long long)timeout, timeout_seq, (long long)(timeout_seq ? *timeout_seq : 0));
 
@@ -1293,7 +1296,6 @@ int rtdm_sem_timeddown(rtdm_sem_t *sem, nanosecs_rel_t timeout,
 
 
 
-
 
 
 	if ((retval = _sem_wait_timed(&sem->sem, timeout, timeout_seq)) == 1) {	
@@ -1327,7 +1329,7 @@ void rtdm_sem_up(rtdm_sem_t *sem)
 {
 
 
-	trace_mark(xn_rtdm_sem_up, "sem %p", sem);
+	trace_mark(xn_rtdm, sem_up, "sem %p", sem);
 
 
 	rt_sem_signal(&sem->sem);
@@ -1359,8 +1361,6 @@ EXPORT_SYMBOL(rtdm_sem_up);
  *
  * @return 0 on success, otherwise:
  *
- * - -EIDRM is returned if @a sem has been destroyed.
- *
  * - -ENOMEM is returned if there is insufficient memory to establish the
  * dynamic binding.
  *
@@ -1388,13 +1388,16 @@ int rtdm_sem_select_bind(rtdm_sem_t *sem, rtdm_selector_t *selector,
 		return -ENOMEM;
 
 	xnlock_get_irqsave(&nklock, s);
-	if (unlikely(sem->sem.magic != RT_SEM_MAGIC))
-		err = -EIDRM;
-	else
-		err = xnselect_bind(&sem->select_block, binding, selector,
-				    type, fd_index, (sem->sem.count > 0));
+	err = xnselect_bind(&sem->select_block, binding, selector,
+			    type, fd_index,
+			    sem->sem.count > 0 ||
+			    sem->sem.magic != RT_SEM_MAGIC);
+
 	xnlock_put_irqrestore(&nklock, s);
 
+	if (err)
+		xnfree(binding);
+
 	return err;
 }
 EXPORT_SYMBOL(rtdm_sem_select_bind);
@@ -1434,6 +1437,7 @@ void rtdm_mutex_init(rtdm_mutex_t *mutex)
 
 
 
+
 	rt_typed_sem_init(mutex, 1, RES_SEM | PRIO_Q);
 }
 
@@ -1549,7 +1553,7 @@ int rtdm_mutex_timedlock(rtdm_mutex_t *mutex, nanosecs_rel_t timeout,
 
 
 
-	trace_mark(xn_rtdm_mutex_timedlock,
+	trace_mark(xn_rtdm, mutex_timedlock,
 		   "mutex %p timeout %Lu timeout_seq %p timeout_seq_value %Lu",
 		   mutex, (long long)timeout, timeout_seq, (long long)(timeout_seq ? *timeout_seq : 0));
 
@@ -1811,7 +1815,7 @@ void rtdm_nrtsig_pend(rtdm_nrtsig_t *nrt_sig);
 
 struct rtdm_mmap_data {
 	void *src_vaddr;
-	unsigned long src_paddr;
+	phys_addr_t src_paddr;
 	struct vm_operations_struct *vm_ops;
 	void *vm_private_data;
 };
@@ -1819,13 +1823,15 @@ struct rtdm_mmap_data {
 static int rtdm_mmap_buffer(struct file *filp, struct vm_area_struct *vma)
 {
 	struct rtdm_mmap_data *mmap_data = filp->private_data;
-	unsigned long vaddr, paddr, maddr, size;
+	unsigned long vaddr, maddr, size;
+	phys_addr_t paddr;
+	int ret;
 
 	vma->vm_ops = mmap_data->vm_ops;
 	vma->vm_private_data = mmap_data->vm_private_data;
 
 	vaddr = (unsigned long)mmap_data->src_vaddr;
-	paddr = (unsigned long)mmap_data->src_paddr;
+	paddr = mmap_data->src_paddr;
 	if (!paddr)
 		/* kmalloc memory */
 		paddr = virt_to_phys((void *)vaddr);
@@ -1849,15 +1855,21 @@ static int rtdm_mmap_buffer(struct file *filp, struct vm_area_struct *vma)
 			vaddr += PAGE_SIZE;
 			mapped_size += PAGE_SIZE;
 		}
-		return 0;
+		xnarch_fault_range(vma);
+		ret = 0;
 	} else
 #endif /* CONFIG_MMU */
 	if (mmap_data->src_paddr)
-		return xnarch_remap_io_page_range(vma, maddr, paddr,
-						  size, PAGE_SHARED);
-	else
-		return 0; /*xnarch_remap_kmem_page_range(vma, maddr, paddr,
-						    size, PAGE_SHARED);*/
+		ret = xnarch_remap_io_page_range(filp, vma, maddr, paddr,
+						 size, PAGE_SHARED);
+	else {
+		ret = xnarch_remap_kmem_page_range(vma, maddr, paddr,
+						   size, PAGE_SHARED);
+		if (!ret)
+			xnarch_fault_range(vma);
+	}
+
+	return ret;
 }
 
 static struct file_operations rtdm_mmap_fops = {
@@ -2021,7 +2033,7 @@ EXPORT_SYMBOL(rtdm_mmap_to_user);
  * Rescheduling: possible.
  */
 int rtdm_iomap_to_user(rtdm_user_info_t *user_info,
-		       unsigned long src_addr, size_t len,
+		       phys_addr_t src_addr, size_t len,
 		       int prot, void **pptr,
 		       struct vm_operations_struct *vm_ops,
 		       void *vm_private_data)
