@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2008 Paolo Mantegazza <mantegazza@aero.polimi.it>
+ * Copyright (C) 2005-2010 Paolo Mantegazza <mantegazza@aero.polimi.it>
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License as
@@ -45,10 +45,10 @@
     } \
 } while(0)
 
-#define RTAI_BUGON(subsystem, cond)  do { \
-    if (unlikely(CONFIG_RTAI_DEBUG_##subsystem > 0 && (cond))) \
-        xnpod_fatal("bug at %s:%d (%s)", __FILE__, __LINE__, (#cond)); \
-} while(0)
+#define RTAI_BUGON(subsystem, cond)  do { /*\
+	if (unlikely(CONFIG_RTAI_DEBUG_##subsystem > 0 && (cond))) \
+		xnpod_fatal("bug at %s:%d (%s)", __FILE__, __LINE__, (#cond)); */ \
+ } while(0)
 
 /* 
   With what above we let some assertion diagnostic. Here below we keep knowledge
@@ -56,8 +56,18 @@
  */
 
 #define xnpod_root_p()          (!current->rtai_tskext(TSKEXT0) || !((RT_TASK *)(current->rtai_tskext(TSKEXT0)))->is_hard)
+#define xnshadow_thread(t)      ((xnthread_t *)current->rtai_tskext(TSKEXT0))
 #define rthal_local_irq_test()  (!rtai_save_flags_irqbit())
 #define rthal_local_irq_enable  rtai_sti 
+#define rthal_domain rtai_domain
+#define rthal_local_irq_disabled()                              \
+({                                                              \
+        unsigned long __flags, __ret;                           \
+        local_irq_save_hw_smp(__flags);                         \
+        __ret = ipipe_test_pipeline_from(&rthal_domain);        \
+        local_irq_restore_hw_smp(__flags);                      \
+        __ret;                                                  \
+})
 
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 
@@ -66,6 +76,8 @@
         static inline void *__check_existence_##name(void) { return &name; } \
         MODULE_PARM(name, "1-" __MODULE_STRING(count) _MODULE_PARM_STRING_##type)
 
+typedef unsigned long phys_addr_t;
+
 #else
 
 #define compat_module_param_array(name, type, count, perm) \
@@ -91,12 +103,27 @@
 typedef unsigned long spl_t;
 typedef struct { volatile unsigned long lock[2]; } xnlock_t;
 
+#ifndef list_first_entry
+#define list_first_entry(ptr, type, member) \
+        list_entry((ptr)->next, type, member)
+#endif
+
+#ifndef local_irq_save_hw_smp
+#ifdef CONFIG_SMP
+#define local_irq_save_hw_smp(flags)    local_irq_save_hw(flags)
+#define local_irq_restore_hw_smp(flags) local_irq_restore_hw(flags)
+#else /* !CONFIG_SMP */
+#define local_irq_save_hw_smp(flags)    do { (void)(flags); } while (0)
+#define local_irq_restore_hw_smp(flags) do { } while (0)
+#endif /* !CONFIG_SMP */
+#endif /* !local_irq_save_hw_smp */
+
 #ifdef CONFIG_SMP
 
-#define DECLARE_XNLOCK(lock)            xnlock_t lock
-#define DECLARE_EXTERN_XNLOCK(lock)     extern xnlock_t lock
-#define DEFINE_XNLOCK(lock)             xnlock_t lock = XNARCH_LOCK_UNLOCKED
-#define DEFINE_PRIVATE_XNLOCK(lock)     static DEFINE_XNLOCK(lock)
+#define DECLARE_XNLOCK(lock)              xnlock_t lock
+#define DECLARE_EXTERN_XNLOCK(lock)       extern xnlock_t lock
+#define DEFINE_XNLOCK(lock)               xnlock_t lock = XNARCH_LOCK_UNLOCKED
+#define DEFINE_PRIVATE_XNLOCK(lock)       static DEFINE_XNLOCK(lock)
 
 static inline void xnlock_init(xnlock_t *lock)
 {
@@ -174,6 +201,7 @@ static inline void xnlock_put_irqrestore(xnlock_t *lock, spl_t flags)
 
 #define xnmalloc  rt_malloc
 #define xnfree    rt_free
+#define xnarch_fault_range(vma)
 
 // in kernel printing (taken from RTDM pet system)
 
@@ -210,7 +238,7 @@ static inline void xnlock_put_irqrestore(xnlock_t *lock, spl_t flags)
 	({ long err = strncpy_from_user(dstP, srcP, n); err; })
 #endif /* CONFIG_M68K */
 
-static inline int xnarch_remap_io_page_range(struct vm_area_struct *vma, unsigned long from, unsigned long to, unsigned long size, pgprot_t prot)
+static inline int xnarch_remap_io_page_range(struct file *filp, struct vm_area_struct *vma, unsigned long from, unsigned long to, unsigned long size, pgprot_t prot)
 {
 #if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
 
@@ -231,6 +259,32 @@ static inline int xnarch_remap_io_page_range(struct vm_area_struct *vma, unsigne
 #endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) */
 }
 
+#define wrap_remap_kmem_page_range(vma,from,to,size,prot) ({ \
+    vma->vm_flags |= VM_RESERVED; \
+    remap_page_range(from,to,size,prot); \
+})
+
+static inline int xnarch_remap_kmem_page_range(struct vm_area_struct *vma, unsigned long from, unsigned long to, unsigned long size, pgprot_t prot)
+{
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
+
+	vma->vm_flags |= VM_RESERVED;
+	return remap_page_range(from, to, size, prot);
+
+#else /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0) */
+
+#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15) && defined(CONFIG_MMU)
+	return remap_pfn_range(vma, from, to >> PAGE_SHIFT, size, prot);
+#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,10)
+	return remap_pfn_range(vma, from, to >> PAGE_SHIFT, size, prot);
+#else /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,10) */
+	vma->vm_flags |= VM_RESERVED;
+	return remap_page_range(from, to, size, prot);
+#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,15) */
+
+#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) */
+}
+
 #include <rtai_shm.h>
 #define __va_to_kva(adr)  UVIRT_TO_KVA(adr)
 
@@ -300,10 +354,10 @@ typedef atomic_t atomic_counter_t;
 
 typedef RTIME xnticks_t;
 
-typedef struct xnstat_runtime {
+typedef struct xnstat_exectime {
         xnticks_t start;
         xnticks_t total;
-} xnstat_runtime_t;
+} xnstat_exectime_t;
 
 typedef struct xnstat_counter {
         int counter;
@@ -322,8 +376,9 @@ typedef struct xnintr {
     xniack_t iack;
     const char *name;
     struct {
-        xnstat_counter_t hits;
-        xnstat_runtime_t account;
+	xnstat_counter_t  hits;
+	xnstat_exectime_t account;
+	xnstat_exectime_t sum;
     } stat[RTAI_NR_CPUS];
 
 } xnintr_t;
@@ -406,6 +461,7 @@ static inline void xntimer_init(xntimer_t *timer, void (*handler)(xntimer_t *))
         memset(timer, 0, sizeof(struct rtdm_timer_struct));
         timer->handler = (void *)handler;
         timer->data    = (unsigned long)timer;
+	timer->next    =  timer->prev = timer;
 }
 
 #define xntimer_set_name(timer, name)
@@ -502,7 +558,7 @@ static inline int emptyq_p(xnqueue_t *queue)
 #define xnthread_test_info    rt_task_test_taskq_retval
 
 #define xnsynch_t                   TASKQ
-#define xnsynch_init                rt_taskq_init
+#define xnsynch_init(s, f, p)       rt_taskq_init(s, f)
 #define xnsynch_destroy             rt_taskq_delete
 #define xnsynch_wakeup_one_sleeper  rt_taskq_ready_one
 #define xnsynch_flush               rt_taskq_ready_all
@@ -520,6 +576,15 @@ static inline void xnsynch_sleep_on(void *synch, xnticks_t timeout, xntmode_t ti
 #define XNSYNCH_FIFO     TASKQ_FIFO
 #define XNSYNCH_RESCHED  1
 
+#define rthal_apc_alloc(name, handler, cookie) \
+	rt_request_srq(nam2num(name), (void *)(handler), NULL);
+
+#define rthal_apc_free(apc) \
+	rt_free_srq((apc))
+
+#define rthal_apc_schedule(apc) \
+	rt_pend_linux_srq((apc))
+
 #ifdef CONFIG_RTAI_RTDM_SELECT
 
 #define SELECT_SIGNAL(select_block, state) \
@@ -538,4 +603,42 @@ do { \
 
 #endif
 
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
+
+#if 1
+#define __WORK_INITIALIZER(n,f,d) {                             \
+        .list   = { &(n).list, &(n).list },                     \
+        .sync = 0,                                              \
+        .routine = (f),                                         \
+        .data = (d),                                            \
+}
+#endif
+
+#define DECLARE_WORK(n,f,d)             struct tq_struct n = __WORK_INITIALIZER(n, f, d)
+#define DECLARE_WORK_NODATA(n, f)       DECLARE_WORK(n, f, NULL)
+#define DECLARE_WORK_FUNC(f)            void f(void *cookie)
+#define DECLARE_DELAYED_WORK_NODATA(n, f) DECLARE_WORK(n, f, NULL)
+
+#define schedule_delayed_work(work, delay) do {                 \
+	if (delay) {                                            \
+		set_current_state(TASK_UNINTERRUPTIBLE);        \
+		schedule_timeout(delay);                        \
+	}                                                       \
+	schedule_task(work);                                    \
+} while (0)
+
+#else
+
+#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,20)
+#define DECLARE_WORK_NODATA(f, n)       DECLARE_WORK(f, n, NULL)
+#define DECLARE_WORK_FUNC(f)            void f(void *cookie)
+#define DECLARE_DELAYED_WORK_NODATA(n, f) DECLARE_DELAYED_WORK(n, f, NULL)
+#else /* >= 2.6.20 */
+#define DECLARE_WORK_NODATA(f, n)       DECLARE_WORK(f, n)
+#define DECLARE_WORK_FUNC(f)            void f(struct work_struct *work)
+#define DECLARE_DELAYED_WORK_NODATA(n, f) DECLARE_DELAYED_WORK(n, f)
+#endif /* >= 2.6.20 */
+
+#endif
+
 #endif /* !_RTAI_XNSTUFF_H */
