@@ -94,10 +94,10 @@ do { \
 /* ++++++++++++++++++++++++++ TASK POINTERS CHECKS +++++++++++++++++++++++++ */
 
 #define CHECK_SENDER_MAGIC(task) \
-do { if (task && task->magic != RT_TASK_MAGIC) return TASK_INVAL; } while (0)
+do { if ((unsigned long)task > RTE_HIGERR && task->magic != RT_TASK_MAGIC) return TASK_INVAL; } while (0)
 
 #define CHECK_RECEIVER_MAGIC(task) \
-do { if (task && task->magic != RT_TASK_MAGIC) return TASK_INVAL; } while (0)
+do { if ((unsigned long)task > RTE_HIGERR && task->magic != RT_TASK_MAGIC) return TASK_INVAL; } while (0)
 
 /* +++++++++++++++++++++++++++++ ASYNC SENDS ++++++++++++++++++++++++++++++++ */
 
@@ -433,7 +433,9 @@ RTAI_SYSCALL_MODE RT_TASK *rt_rpc(RT_TASK *task, unsigned long to_do, void *resu
 	rt_current->msg_queue.task = task;
 	RT_SCHEDULE_BOTH(task, cpuid);
 	if (rt_current->msg_queue.task == rt_current) {
-		*(unsigned long *)result = rt_current->msg;
+		if (result) {
+			*(unsigned long *)result = rt_current->msg;
+		}
 	} else {
 		msg_not_sent();
 	}
@@ -507,7 +509,9 @@ RTAI_SYSCALL_MODE RT_TASK *rt_rpc_if(RT_TASK *task, unsigned long to_do, void *r
 		rt_current->msg_queue.task = task;
 		RT_SCHEDULE_BOTH(task, cpuid);
 		if (rt_current->msg_queue.task == rt_current) {
-			*(unsigned long *)result = rt_current->msg;
+			if (result) {
+				*(unsigned long *)result = rt_current->msg;
+			}
 		} else {
 			msg_not_sent();
 		}
@@ -593,7 +597,9 @@ RTAI_SYSCALL_MODE RT_TASK *rt_rpc_until(RT_TASK *task, unsigned long to_do, void
 	enq_timed_task(rt_current);
 	RT_SCHEDULE_BOTH(task, cpuid);
 	if (rt_current->msg_queue.task == rt_current) {
-		*(unsigned long *)result = rt_current->msg;
+		if (result) {
+			*(unsigned long *)result = rt_current->msg;
+		}
 	} else {
 		msg_not_sent();
 	}
@@ -769,7 +775,9 @@ RTAI_SYSCALL_MODE RT_TASK *rt_evdrp(RT_TASK *task, void *msg)
 	ASSIGN_RT_CURRENT;
 	if (!task) task = (rt_current->msg_queue.next)->task;
 	if ((task->state & (RT_SCHED_SEND | RT_SCHED_RPC)) && task->msg_queue.task == rt_current) {
-		*(unsigned long *)msg = task->msg;
+		if (msg) {
+			*(unsigned long *)msg = task->msg;
+		}
 	} else {
 		task = NULL;
 	}
@@ -825,7 +833,9 @@ RTAI_SYSCALL_MODE RT_TASK *rt_receive(RT_TASK *task, void *msg)
 	if ((task->state & (RT_SCHED_SEND | RT_SCHED_RPC)) && task->msg_queue.task == rt_current) {
 		dequeue_blocked(task);
 		rem_timed_task(task);
-		*(unsigned long *)msg = task->msg;
+		if (msg) {
+			*(unsigned long *)msg = task->msg;
+		}
 		rt_current->msg_queue.task = task;
 		if (task->state & RT_SCHED_SEND) {
 			int sched;
@@ -849,7 +859,9 @@ RTAI_SYSCALL_MODE RT_TASK *rt_receive(RT_TASK *task, void *msg)
 		rem_ready_current(rt_current);
 		rt_current->msg_queue.task = task != rt_current ? task : (RT_TASK *)0;
 		rt_schedule();
-		*(unsigned long *)msg = rt_current->msg;
+		if (msg) {
+			*(unsigned long *)msg = rt_current->msg;
+		}
 	}
 	if (rt_current->ret_queue.task) {
 		msg_not_received();
@@ -911,7 +923,9 @@ RTAI_SYSCALL_MODE RT_TASK *rt_receive_if(RT_TASK *task, void *msg)
 	if ((task->state & (RT_SCHED_SEND | RT_SCHED_RPC)) && task->msg_queue.task == rt_current) {
 		dequeue_blocked(task);
 		rem_timed_task(task);
-		*(unsigned long *)msg = task->msg;
+		if (msg) {
+			*(unsigned long *)msg = task->msg;
+		}
 		rt_current->msg_queue.task = task;
 		if (task->state & RT_SCHED_SEND) {
 			int sched;
@@ -1003,7 +1017,9 @@ RTAI_SYSCALL_MODE RT_TASK *rt_receive_until(RT_TASK *task, void *msg, RTIME time
 	if ((task->state & (RT_SCHED_SEND | RT_SCHED_RPC)) && task->msg_queue.task == rt_current) {
 		dequeue_blocked(task);
 		rem_timed_task(task);
-		*(unsigned long *)msg = task->msg;
+		if (msg) {
+			*(unsigned long *)msg = task->msg;
+		}
 		rt_current->msg_queue.task = task;
 		if (task->state & RT_SCHED_SEND) {
 			int sched;
@@ -1029,7 +1045,9 @@ RTAI_SYSCALL_MODE RT_TASK *rt_receive_until(RT_TASK *task, void *msg, RTIME time
 			rt_current->msg_queue.task = task != rt_current ? task : (RT_TASK *)0;
 			enq_timed_task(rt_current);
 			rt_schedule();
-			*(unsigned long *)msg = rt_current->msg;
+			if (msg) {
+				*(unsigned long *)msg = rt_current->msg;
+			}
 		}
 	}
 	if (rt_current->ret_queue.task) {
@@ -1148,8 +1166,9 @@ RTAI_SYSCALL_MODE RT_TASK *rt_rpcx(RT_TASK *task, void *smsg, void *rmsg, int ss
 {
 	if (task) {
 		struct mcb_t mcb;
+		unsigned long retrep;
 		SET_RPC_MCB();
-		return rt_rpc(task, (unsigned long)&mcb, &mcb.rbytes);
+		return rt_rpc(task, (unsigned long)&mcb, &retrep);
 	}
 	return 0;
 }
@@ -1197,8 +1216,9 @@ RTAI_SYSCALL_MODE RT_TASK *rt_rpcx_if(RT_TASK *task, void *smsg, void *rmsg, int
 {
 	if (task) {
 		struct mcb_t mcb;
+		unsigned long retrep;
 		SET_RPC_MCB();
-		return rt_rpc_if(task, (unsigned long)&mcb, &mcb.rbytes);
+		return rt_rpc_if(task, (unsigned long)&mcb, &retrep);
 	}
 	return 0;
 }
@@ -1251,8 +1271,9 @@ RTAI_SYSCALL_MODE RT_TASK *rt_rpcx_until(RT_TASK *task, void *smsg, void *rmsg,
 {
 	if (task) {
 		struct mcb_t mcb;
+		unsigned long retrep;
 		SET_RPC_MCB();
-		return rt_rpc_until(task, (unsigned long)&mcb, &mcb.rbytes, time);
+		return rt_rpc_until(task, (unsigned long)&mcb, &retrep, time);
 	}
 	return 0;
 }
@@ -1305,19 +1326,20 @@ RTAI_SYSCALL_MODE RT_TASK *rt_rpcx_timed(RT_TASK *task, void *smsg, void *rmsg,
 {
 	if (task) {
 		struct mcb_t mcb;
+		unsigned long retrep;
 		SET_RPC_MCB();
-		return rt_rpc_timed(task, (unsigned long)&mcb, &mcb.rbytes, delay);
+		return rt_rpc_timed(task, (unsigned long)&mcb, &retrep, delay);
 	}
 	return 0;
 }
 
-#define task_mcb (task->mcb)
+#define SEND_RCV_BYTES 0xFFFFFFFF
 #define SET_SEND_MCB() \
 	do { \
-		task_mcb.sbuf   = msg; \
-		task_mcb.sbytes = size; \
-		task_mcb.rbuf   = 0; \
-		task_mcb.rbytes = 0; \
+		mcb.sbuf   = msg; \
+		mcb.sbytes = size; \
+		mcb.rbuf   = NULL; \
+		mcb.rbytes = SEND_RCV_BYTES; \
 	} while (0)
 
 /**
@@ -1356,8 +1378,10 @@ RTAI_SYSCALL_MODE RT_TASK *rt_rpcx_timed(RT_TASK *task, void *smsg, void *rmsg,
 RTAI_SYSCALL_MODE RT_TASK *rt_sendx(RT_TASK *task, void *msg, int size) 
 {
 	if (task) {
+		struct mcb_t mcb;
+		unsigned long retrep;
 		SET_SEND_MCB();
-		return rt_send(task, (unsigned long)&task_mcb);
+		return rt_rpc(task, (unsigned long)&mcb, &retrep);
 	}
 	return 0;
 }
@@ -1396,11 +1420,34 @@ RTAI_SYSCALL_MODE RT_TASK *rt_sendx(RT_TASK *task, void *msg, int size)
  */
 RTAI_SYSCALL_MODE RT_TASK *rt_sendx_if(RT_TASK *task, void *msg, int size)
 {
-	if (task) {
-		SET_SEND_MCB();
-		return rt_send_if(task, (unsigned long)&task_mcb);
+	DECLARE_RT_CURRENT;
+	unsigned long flags;
+
+	CHECK_SENDER_MAGIC(task);
+
+	flags = rt_global_save_flags_and_cli();
+	ASSIGN_RT_CURRENT;
+	if ((task->state & RT_SCHED_RECEIVE) &&
+	      (!task->msg_queue.task || task->msg_queue.task == rt_current)) {
+		task->mcb.sbuf = msg;
+		task->mcb.sbytes = size;
+		task->msg = (unsigned long)&task->mcb;
+		task->msg_queue.task = rt_current;
+		task->ret_queue.task = NULL;
+		rem_timed_task(task);
+		if (task->state != RT_SCHED_READY && (task->state &= ~(RT_SCHED_RECEIVE | RT_SCHED_DELAYED)) == RT_SCHED_READY) {
+			enq_ready_task(task);
+			RT_SCHEDULE(task, cpuid);
+		}
+		if (rt_current->msg_queue.task != rt_current) {
+			rt_current->msg_queue.task = rt_current;
+			task = NULL;
+		}
+	} else {
+		task = NULL;
 	}
-	return 0;
+	rt_global_restore_flags(flags);
+	return task;
 }
 
 
@@ -1448,8 +1495,10 @@ RTAI_SYSCALL_MODE RT_TASK *rt_sendx_if(RT_TASK *task, void *msg, int size)
 RTAI_SYSCALL_MODE RT_TASK *rt_sendx_until(RT_TASK *task, void *msg, int size, RTIME time)
 {
 	if (task) {
+		struct mcb_t mcb;
+		unsigned long retrep;
 		SET_SEND_MCB();
-		return rt_send_until(task, (unsigned long)&task_mcb, time);
+		return rt_rpc_until(task, (unsigned long)&mcb, &retrep, time);
 	}
 	return 0;
 }
@@ -1499,8 +1548,10 @@ RTAI_SYSCALL_MODE RT_TASK *rt_sendx_until(RT_TASK *task, void *msg, int size, RT
 RTAI_SYSCALL_MODE RT_TASK *rt_sendx_timed(RT_TASK *task, void *msg, int size, RTIME delay)
 {
 	if (task) {
+		struct mcb_t mcb;
+		unsigned long retrep;
 		SET_SEND_MCB();
-		return rt_send_timed(task, (unsigned long)&task_mcb, delay);
+		return rt_rpc_timed(task, (unsigned long)&mcb, &retrep, delay);
 	}
 	return 0;
 }
@@ -1555,7 +1606,7 @@ RTAI_SYSCALL_MODE RT_TASK *rt_returnx(RT_TASK *task, void *msg, int size)
 		if ((mcb = (struct mcb_t *)task->msg)->rbytes < size) {
 			size = mcb->rbytes;
 		}
-		if (size) {
+		if (msg && size > 0) {
 			memcpy(mcb->rbuf, msg, size);
 		}
 		_rt_return(0UL);
@@ -1568,9 +1619,12 @@ RTAI_SYSCALL_MODE RT_TASK *rt_returnx(RT_TASK *task, void *msg, int size)
 
 #define DO_RCV_MSG() \
 	do { \
-		if ((*len = size <= mcb->sbytes ? size : mcb->sbytes)) { \
+		if (msg && (*len = size <= mcb->sbytes ? size : mcb->sbytes)) { \
 			memcpy(msg, mcb->sbuf, *len); \
 		} \
+		if ((unsigned long)task > RTE_HIGERR && !mcb->rbuf && mcb->rbytes == SEND_RCV_BYTES) { \
+			rt_return(task, 0UL); \
+		} \
 	} while (0)
 
 
@@ -1609,11 +1663,10 @@ RTAI_SYSCALL_MODE RT_TASK *rt_returnx(RT_TASK *task, void *msg, int size)
 RTAI_SYSCALL_MODE RT_TASK *rt_evdrpx(RT_TASK *task, void *msg, int size, long *len)
 {
 	struct mcb_t *mcb;
-	if ((task = rt_evdrp(task, (unsigned long *)&mcb))) {
+	if ((unsigned long)(task = rt_evdrp(task, (unsigned long *)&mcb)) > RTE_HIGERR) {
 		DO_RCV_MSG();
-		return task;
 	}
-	return 0;
+	return task;
 }
 
 
@@ -1660,11 +1713,10 @@ RTAI_SYSCALL_MODE RT_TASK *rt_evdrpx(RT_TASK *task, void *msg, int size, long *l
 RTAI_SYSCALL_MODE RT_TASK *rt_receivex(RT_TASK *task, void *msg, int size, long *len)
 {
 	struct mcb_t *mcb;
-	if ((task = rt_receive(task, (unsigned long *)&mcb))) {
+	if ((unsigned long)(task = rt_receive(task, (unsigned long *)&mcb)) > RTE_HIGERR) {
 		DO_RCV_MSG();
-		return task;
 	}
-	return 0;
+	return task;
 }
 
 
@@ -1711,11 +1763,10 @@ RTAI_SYSCALL_MODE RT_TASK *rt_receivex(RT_TASK *task, void *msg, int size, long
 RTAI_SYSCALL_MODE RT_TASK *rt_receivex_if(RT_TASK *task, void *msg, int size, long *len)
 {
 	struct mcb_t *mcb;
-	if ((task = rt_receive_if(task, (unsigned long *)&mcb))) {
+	if ((unsigned long)(task = rt_receive_if(task, (unsigned long *)&mcb)) > RTE_HIGERR) {
 		DO_RCV_MSG();
-		return task;
 	}
-	return 0;
+	return task;
 }
 
 
@@ -1768,11 +1819,10 @@ RTAI_SYSCALL_MODE RT_TASK *rt_receivex_if(RT_TASK *task, void *msg, int size, lo
 RTAI_SYSCALL_MODE RT_TASK *rt_receivex_until(RT_TASK *task, void *msg, int size, long *len, RTIME time)
 {
 	struct mcb_t *mcb;
-	if ((task = rt_receive_until(task, (unsigned long *)&mcb, time))) {
+	if ((unsigned long)(task = rt_receive_until(task, (unsigned long *)&mcb, time)) > RTE_HIGERR) {
 		DO_RCV_MSG();
-		return task;
 	}
-	return 0;
+	return task;
 }
 
 
@@ -1825,11 +1875,10 @@ RTAI_SYSCALL_MODE RT_TASK *rt_receivex_until(RT_TASK *task, void *msg, int size,
 RTAI_SYSCALL_MODE RT_TASK *rt_receivex_timed(RT_TASK *task, void *msg, int size, long *len, RTIME delay)
 {
 	struct mcb_t *mcb;
-	if ((task = rt_receive_timed(task, (unsigned long *)&mcb, delay))) {
+	if ((unsigned long)(task = rt_receive_timed(task, (unsigned long *)&mcb, delay)) > RTE_HIGERR) {
 		DO_RCV_MSG();
-		return task;
 	}
-	return 0;
+	return task;
 }
 
 /* +++++++++++++++++++++++++++++++ PROXIES ++++++++++++++++++++++++++++++++++ */
