@@ -1,5 +1,6 @@
 #ifndef __SCICOS_BLOCK4_H__
 #define __SCICOS_BLOCK4_H__
+#include <float.h>
 
 #ifdef __STDC__
 	#include <stdlib.h>
@@ -60,6 +61,9 @@ typedef struct {
    void **work;
    int nmode;
    int *mode;
+   double *alpha;
+   double *beta;
+
  } scicos_block;
 
 
@@ -100,6 +104,8 @@ typedef struct {
 */
 #define ReInitialization 6
 
+typedef enum { PHASE_MESHPOINT=0, PHASE_DISCRETE=1, PHASE_TRY_MFX=2 } PHASE_SIMULATOR;
+
 #define DoColdRestart(block)        (do_cold_restart()) 
 #define GetSimulationPhase(block)   (get_phase_simulation())
 #define GetScicosTime(block)        (get_scicos_time())
@@ -109,9 +115,15 @@ typedef struct {
 #define GetBlockError(block)        (get_block_error())
 #define SetBlockError(block,val)    (set_block_error(val))
 #define StopSimulation(block,val)   (end_scicos_sim()) 
+#define IsHotReStart(block)         (what_is_hot())
+#define isinTryPhase(block)         ( GetSimulationPhase(block)==PHASE_TRY_MFX  )
+#define areModesFixed(block)        ( GetSimulationPhase(block)==PHASE_TRY_MFX )
+#define isatMeshPoint(block)        ( GetSimulationPhase(block)==PHASE_MESHPOINT )
+#define GetSQRU(block)              (sqrt(DBL_EPSILON))
 
 /* utility function for block declaration */
 void do_cold_restart();
+int what_is_hot();
 int get_phase_simulation();
 int get_fcaller_id();
 double get_scicos_time();
@@ -131,6 +143,8 @@ double Get_Jacobian_ci(void);
 double Get_Scicos_SQUR(void);
 void Set_Jacobian_flag(int flag);
 
+#define SetAjac(blk,n) (Set_Jacobian_flag( n))
+
 int Convert_number (char *, double *);
 void homotopy(double *);
 int hfjac_(double *, double *, int *);
@@ -178,12 +192,12 @@ extern int s_cmp();
 #define SCSINT_COP int
 #define SCSINT8_COP char
 #define SCSINT16_COP short
-#define SCSINT32_COP long
+#define SCSINT32_COP int
 #define SCSUINT_COP unsigned int
 #define SCSUINT8_COP unsigned char
 #define SCSUINT16_COP unsigned short
-#define SCSUINT32_COP unsigned long
-#define SCSBOOL_COP long
+#define SCSUINT32_COP unsigned int
+#define SCSBOOL_COP int
 #define SCSUNKNOW_COP double
 
  /* scicos_block macros definition :
@@ -294,6 +308,10 @@ extern int s_cmp();
 /**
    \brief Get regular output port pointer of port number x.
 */
+
+#define GetOutPtrs(blk) (blk->outptr)
+#define GetInPtrs(blk) (blk->inptr)
+
 #define GetOutPortPtrs(blk,x) ((((x)>0)&((x)<=(blk->nout))) ? (blk->outptr[x-1]) : NULL)
 
 /**
@@ -482,6 +500,10 @@ extern int s_cmp();
    \brief Get pointer of continuous state properties register.
 */
 #define GetXpropPtrs(blk) (blk->xprop)
+#define GetXpropPtrs(blk) (blk->xprop)
+#define GetAlphaPt(blk)  (blk->alpha)
+#define GetBetaPt(blk)  (blk->beta)
+
 
 /**
    \brief Get number of discrete state.
@@ -741,4 +763,6 @@ extern int s_cmp();
 #endif
 
 
+
+
 #endif /* __SCICOS_BLOCK_H__ */
